"use strict";
var filters = require('morphologicalOperations');

// Note that, conceptually, we can think of the positive part of the y-axis pointing up (as is common in mathematics), but still process the image "as usual" (positive y going down).
// Essentially, images are simply considered to be displayed upside down.

var numFrameAngles = 11, angleStep = Math.PI/numFrameAngles;

exports.filter = function (imageData, filterSpec, projectData, progressCallback){
    var inputData, inputDataOrg, imageDX, imageDY, imageST, liftedData, imageTraceVal, imageTraceValMax;
    var width = imageData.width;
    var height = imageData.height;
    var x, y, c, ry, p, j, k, filterLength, angle, v, filterFuncH, filterFuncV, filterFunc, curLength, curOffset, filter;
    var inv255255 = 1/(255*255);
    var filterParts, filterName, filterLength;
    
    if (!filterSpec || filterSpec.length==0) {
        filter = false;
    } else if (typeof filterSpec=="string" || filterSpec instanceof String) {
        filter = true;

        filterParts = filterSpec.split('-');
        filterName = filterParts[0];
        filterLength = parseInt(filterParts[1], 10);
        
        filterFuncH = filters['horizontal' + filterName];
        filterFuncV = filters['vertical' + filterName];
        if (filterFuncH===undefined || !(filterFuncH instanceof Function) || filterFuncV===undefined || !(filterFuncV instanceof Function)) {
            throw "Unsupported filter!";
            return;
        }
    } else {
        throw "Invalid filter specification!";
    }
    
    if (width<=1 || height <=1) return; // Nothing interesting to do. TODO: Handle more gracefully? (The reason I need it is that below I assume that we never have x==0 && x+1==width.)
    
    // Convert image to floating-point RGB (instead of RGBA).
    inputDataOrg = imageData.data;
    
    inputData = new Float32Array(width*height*3);
    for(y=0; y<height; y++) {
        for(x=0; x<width; x++) {
            p = x+width*y;
            inputData[3*p+0] = inv255255*(inputDataOrg[4*p+3]*inputDataOrg[4*p+0] + (255-inputDataOrg[4*p+3])*255);
            inputData[3*p+1] = inv255255*(inputDataOrg[4*p+3]*inputDataOrg[4*p+1] + (255-inputDataOrg[4*p+3])*255);
            inputData[3*p+2] = inv255255*(inputDataOrg[4*p+3]*inputDataOrg[4*p+2] + (255-inputDataOrg[4*p+3])*255);
        }
    }
    
    // Compute partial derivatives
    imageDX = new Float32Array(inputData.length);
    for(y=0; y<height; y++) {
        for(x=0; x<width; x++) {
            if (x==0) {
                for(c=0; c<3; c++) {
                    imageDX[c+3*(x+width*y)] = inputData[c+3*(x+1+width*y)] - inputData[c+3*(x+width*y)];
                }
            } else if (x+1<width) {
                for(c=0; c<3; c++) {
                    imageDX[c+3*(x+width*y)] = (inputData[c+3*(x+1+width*y)] - inputData[c+3*(x-1+width*y)])/2;
                }
            } else {
                for(c=0; c<3; c++) {
                    imageDX[c+3*(x+width*y)] = inputData[c+3*(x+width*y)] - inputData[c+3*(x-1+width*y)];
                }
            }
        }
    }

    imageDY = new Float32Array(inputData.length);
    for(y=0; y<height; y++) {
        if (y==0) {
            for(x=0; x<width; x++) {
                for(c=0; c<3; c++) {
                    imageDY[c+3*(x+width*y)] = inputData[c+3*(x+width*(y+1))] - inputData[c+3*(x+width*y)];
                }
            }
        } else if (y+1<height) {
            for(x=0; x<width; x++) {
                for(c=0; c<3; c++) {
                    imageDY[c+3*(x+width*y)] = (inputData[c+3*(x+width*(y+1))] - inputData[c+3*(x+width*(y-1))])/2;
                }
            }
        } else {
            for(x=0; x<width; x++) {
                for(c=0; c<3; c++) {
                    imageDY[c+3*(x+width*y)] = inputData[c+3*(x+width*y)] - inputData[c+3*(x+width*(y-1))];
                }
            }
        }
    }
    
    // Compute structure tensor
    imageST = new Float32Array(width*height*3);
    for(y=0; y<height; y++) {
        for(x=0; x<width; x++) {
            imageST[0+3*(x+width*y)] = imageDX[0+3*(x+width*y)]*imageDX[0+3*(x+width*y)] + imageDX[1+3*(x+width*y)]*imageDX[1+3*(x+width*y)] + imageDX[2+3*(x+width*y)]*imageDX[2+3*(x+width*y)]
            imageST[1+3*(x+width*y)] = imageDY[0+3*(x+width*y)]*imageDY[0+3*(x+width*y)] + imageDY[1+3*(x+width*y)]*imageDY[1+3*(x+width*y)] + imageDY[2+3*(x+width*y)]*imageDY[2+3*(x+width*y)]
            imageST[2+3*(x+width*y)] = imageDX[0+3*(x+width*y)]*imageDY[0+3*(x+width*y)] + imageDX[1+3*(x+width*y)]*imageDY[1+3*(x+width*y)] + imageDX[2+3*(x+width*y)]*imageDY[2+3*(x+width*y)]
            imageST[2+3*(x+width*y)] *= Math.sqrt(2);
        }
    }

    // Filter
    if (filter) {
        liftedData = lift(imageST);
        if (progressCallback) progressCallback(0); // Initial status "update"
        for(j=0; j<numFrameAngles; j++) {
            for(k=0; k<3; k++) {
                // The angle is the angle associated with the current tensor component.
                angle = -j*angleStep+k*(Math.PI/3) + Math.PI/2; // By convention we use the "short" angles.
                v = [Math.abs(Math.cos(angle)), Math.abs(Math.sin(angle))];
                if (v[0]>=v[1]) {
                    filterFunc = filterFuncH;
                    curLength = v[0]*filterLength;
                    curOffset = -Math.tan(angle); // Use shift opposite to the shift of the direction vector for the tensor component.
                } else {
                    filterFunc = filterFuncV;
                    curLength = v[1]*filterLength;
                    curOffset = -Math.tan(Math.PI/2-angle);
                }
                filterFunc(liftedData.subarray((j*3+k)*width*height,(j*3+k+1)*width*height), width,height, curLength, curOffset);
            
                // Status update
                if (progressCallback) progressCallback((j*3+k+1)/(numFrameAngles*3));
            }
        }
        if (projectData) imageST = project(liftedData);
    }

    // Compute output
    imageTraceVal = new Float32Array(width*height);
    imageTraceValMax = 0;
    for(y=0; y<height; y++) {
        for(x=0; x<width; x++) {
            if (!filter || projectData) {
                // Use the square root of trace of tensor (without filtering this is equivalent to taking the norm of the gradient).
                imageTraceVal[x+width*y] = Math.sqrt(imageST[0+3*(x+width*y)] + imageST[1+3*(x+width*y)]);
            } else {
                // Use the square root of the maximum of the frame coefficients (without filtering this is, again, equivalent to taking the norm of the gradient).
                for(j=0; j<numFrameAngles; j++) {
                    for(k=0; k<3; k++) {
                        imageTraceVal[x+width*y] = Math.max(imageTraceVal[x+width*y], liftedData[x+width*(y+height*(k+3*j))]);
                    }
                }
                imageTraceVal[x+width*y] = Math.sqrt(imageTraceVal[x+width*y]);
            }
            imageTraceValMax = Math.max(imageTraceValMax, imageTraceVal[x+width*y]);
        }
    }
    for(y=0; y<height; y++) {
        for(x=0; x<width; x++) {
            inputDataOrg[0+4*(x+width*y)] = imageTraceVal[x+width*y]*(255/imageTraceValMax) + 0.5;
            inputDataOrg[1+4*(x+width*y)] = imageTraceVal[x+width*y]*(255/imageTraceValMax) + 0.5;
            inputDataOrg[2+4*(x+width*y)] = imageTraceVal[x+width*y]*(255/imageTraceValMax) + 0.5;
            inputDataOrg[3+4*(x+width*y)] = 255;
        }
    }
}

/*function rotate(data, angle) {
    var newData = new Float32Array(data.length);
    var numPixels = data.length/3;
    var coefficients = new Float32Array(3);
    var i, j;
    
    coefficients[0] = Math.pow(Math.cos(angle),2);
    coefficients[1] = Math.pow(Math.sin(angle),2);
    coefficients[2] = Math.sqrt(2)*Math.cos(angle)*Math.sin(angle);
    
    for(i=0; i<numPixels; i++) {
        newData[3*i+0] = coefficients[0]*data[3*i+0] + coefficients[1]*data[3*i+1] - coefficients[2]*data[3*i+2];
        newData[3*i+1] = coefficients[1]*data[3*i+0] + coefficients[0]*data[3*i+1] + coefficients[2]*data[3*i+2];
        newData[3*i+2] = coefficients[2]*(data[3*i+0]-data[3*i+1]) + (coefficients[0]-coefficients[1])*data[3*i+2];
    }
    
    return newData;
}*/

function lift(data) {
    var angle;
    var newData = new Float32Array(numFrameAngles*data.length);
    var numPixels = data.length/3;
    var coefficients = new Float32Array(3*numFrameAngles);
    var i, j;
    
    for(j=0; j<numFrameAngles; j++) {
        angle = j*angleStep;
        coefficients[j*3+0] = Math.pow(Math.cos(angle),2);
        coefficients[j*3+1] = Math.pow(Math.sin(angle),2);
        coefficients[j*3+2] = Math.sqrt(2)*Math.cos(angle)*Math.sin(angle);
    }
    
    var newData0, newData1, newData2;
    for(i=0; i<numPixels; i++) {
        for(j=0; j<numFrameAngles*3; j+=3) {
            newData0 = coefficients[j+0]*data[3*i+0] + coefficients[j+1]*data[3*i+1] - coefficients[j+2]*data[3*i+2];
            newData1 = coefficients[j+1]*data[3*i+0] + coefficients[j+0]*data[3*i+1] + coefficients[j+2]*data[3*i+2];
            newData2 = coefficients[j+2]*(data[3*i+0]-data[3*i+1]) + (coefficients[j+0]-coefficients[j+1])*data[3*i+2];
            newData[i+(j+0)*numPixels] = newData0;
            newData[i+(j+1)*numPixels] = (newData0 + 3*newData1 + Math.sqrt(6)*newData2)/4;
            newData[i+(j+2)*numPixels] = (newData0 + 3*newData1 - Math.sqrt(6)*newData2)/4;
            /* This uses the dual basis.
            newData[i+(j+0)*numPixels] = newData0 - newData1/3;
            newData[i+(j+1)*numPixels] = (2/3.0)*newData1 + Math.sqrt(2/3.0)*newData2;
            newData[i+(j+2)*numPixels] = (2/3.0)*newData1 - Math.sqrt(2/3.0)*newData2;*/
        }
    }
    
    return newData;
}

function project(data) {
    var angle;
    var newData = new Float32Array(data.length/numFrameAngles);
    var numPixels = newData.length/3;
    var coefficients = new Float32Array(3*numFrameAngles);
    var i, j;

    for(j=0; j<numFrameAngles; j++) {
        // Using the inverse angle.
        // Note that I pre-divide by the number of angles.
        angle = -j*angleStep;
        coefficients[j*3+0] = Math.pow(Math.cos(angle),2)/numFrameAngles;
        coefficients[j*3+1] = Math.pow(Math.sin(angle),2)/numFrameAngles;
        coefficients[j*3+2] = Math.sqrt(2)*Math.cos(angle)*Math.sin(angle)/numFrameAngles;
    }
    
    var newData0, newData1, newData2;
    for(i=0; i<numPixels; i++) {
        for(j=0; j<numFrameAngles*3; j+=3) {
            newData0 = data[i+(j+0)*numPixels];
            newData1 = (2*(data[i+(j+1)*numPixels]+data[i+(j+2)*numPixels]) - data[i+(j+0)*numPixels])/3;
            newData2 = Math.sqrt(2/3.0)*(data[i+(j+1)*numPixels] - data[i+(j+2)*numPixels]);
            /* This is based on the dual basis.
            newData0 = data[i+(j+0)*numPixels] + (data[i+(j+1)*numPixels] + data[i+(j+2)*numPixels])/4;
            newData1 = (3/4.0)*(data[i+(j+1)*numPixels] + data[i+(j+2)*numPixels]);
            newData2 = Math.sqrt(3/8.0)*(data[i+(j+1)*numPixels] - data[i+(j+2)*numPixels]);*/
            newData[3*i+0] += coefficients[j+0]*newData0 + coefficients[j+1]*newData1 - coefficients[j+2]*newData2;
            newData[3*i+1] += coefficients[j+1]*newData0 + coefficients[j+0]*newData1 + coefficients[j+2]*newData2;
            newData[3*i+2] += coefficients[j+2]*(newData0-newData1) + (coefficients[j+0]-coefficients[j+1])*newData2;
        }
    }
    
    return newData;
}

