"use strict";

function lessThanOrEqualTo(a,b) { return a <= b; }
function greaterThanOrEqualTo(a,b) { return a >= b; }

function lerp(a, b, f) {
    return a + f*(b-a);
}

// These methods are based on:
//   Computationally efficient, one-pass algorithm for morphological filters Journal of Visual Communication and Image Representation, Vol. 22, No. 5. (July 2011), pp. 411-420, doi:10.1016/j.jvcir.2011.03.005 by Petr Dokládal, Eva Dokládalová

// This is the "class" used to implement the required queue, the actual operators follow below.
function MorphologicalQueue(size,cmp) {
    var l2 = 0, s = 1;
    while(l2<31 && s<size) { // Compute smallest power-of-two greater than or equal to "size".
        l2++;
        s = s<<1;
    }
    this.size = size>>0; // Truncate size, essentially a hint to the compiler that size is supposed to be an integer.
    this.queue = new Array(2<<l2); // Two elements per item (a position and a value). The array is used circularly, as there are never more than "size" items in the queue. We use a power-of-two to make the modulo operation easier.
    this.queueBegin = 0; // These indices are /not/ used circularly, as this makes the logic easier. So instead of keeping these indices within the bounds of the array, every access to the array is guarded.
    this.queueEnd = 0;
    this.cmp = cmp;
}
MorphologicalQueue.prototype.firstVal = function () {
    return this.queueBegin<this.queueEnd ? this.queue[(this.queueBegin+1)&(this.queue.length-1)] : undefined;
};
MorphologicalQueue.prototype.empty = function () {
    this.queueBegin = 0;
    this.queueEnd = 0;
};
MorphologicalQueue.prototype.enqueue = function (pos, val) {
    var q = this.queue, cmp = this.cmp, lengthMask = q.length-1;

    // Unshift first two elements if they are falling off the window.
    if (this.queueBegin<this.queueEnd && pos >= q[this.queueBegin&lengthMask]) {
        this.queueBegin += 2;
    }

    if (val!==undefined) {
        // Pop elements as long as they are "useless".
        while(this.queueBegin<this.queueEnd && cmp(val,q[(this.queueEnd+lengthMask)&lengthMask])) { // (end+lengthMask)&lengthMask = (end-1) modulo q.length
            this.queueEnd -= 2;
        }
        
        // Push new element.
        q[(this.queueEnd++)&lengthMask] = pos+this.size;
        q[(this.queueEnd++)&lengthMask] = val;
    }
};

// The following operators are available:
//   horizontalDilation, -Erosion, -Closing, -Opening, -OpenClose, -CloseOpen: dilation (etc.) by a horizontal line segment
//   verticalDilation, -Erosion, -Closing, -Opening, -OpenClose, -CloseOpen: dilation (etc.) by a vertical line segment
//   erosion-, dilation-, open-, close-, occoSquare: erosion (etc.) by a square
// The operators support the following arguments:
//   data: array with data (row-major, each "pixel" only has a single (grey) value)
//   width, height: size of image
//   length: total length of one side of the square
//   skewOffset: if given and non-zero, effectively skews the line segment (only for the operators using a line segment)
// The length parameter is rounded to the nearest odd integer for erosionSquare and dilationSquare, such that even integers are rounded up.
// For the opening, closing and occo operators the length parameter is rounded to the smallest integer strictly larger than the given length to determine the actual pixel size of the SE.
// The motivation for these choices is that a SE with a length of 2 pixels touches three pixel centers.

exports.horizontalDilation = function (data, width, height, lengthFrac, skewOffset) {
    if (skewOffset===undefined) skewOffset = 0;
    var x, y, c, ly, uy, fy, sv;
    var radius = Math.floor(lengthFrac/2);
    var queue = new MorphologicalQueue(2*radius+1, greaterThanOrEqualTo);
    var skewedBeginY = Math.min(0,Math.floor((width-1)*skewOffset));
    var skewedEndY = height + Math.max(0,Math.ceil((width-1)*skewOffset));
    var newData = new Float32Array(width*(height+1));
    for(y=skewedBeginY; y<skewedEndY; y++) {
        queue.empty();
        for(x=0; x<width+radius; x++) {
            fy = y - x*skewOffset;
            ly = Math.floor(fy);
            uy = Math.ceil(fy);
            fy -= ly;
            if (x<width && 0<=uy && ly<height) {
                if (ly<0 || uy==ly) {
                    sv = data[uy*width+x];
                } else if (uy==height) {
                    sv = data[ly*width+x];
                } else {
                    sv = data[ly*width+x] + fy*(data[uy*width+x]-data[ly*width+x]);
                }
                queue.enqueue(x, sv);
            } else {
                queue.enqueue(x);
            }
            if (x>=radius) {
                fy = y - (x-radius)*skewOffset;
                ly = Math.floor(fy);
                uy = Math.ceil(fy);
                fy -= ly;
                if (0<=uy && ly<height) {
                    newData[uy*width+x-radius] = queue.firstVal();
                }
            }
        }
    }
    for(y=0; y<height; y++) {
        for(x=0; x<width; x++) {
            fy = y - x*skewOffset;
            ly = Math.floor(fy);
            uy = Math.ceil(fy);
            fy -= ly;
            data[y*width+x] = fy*newData[y*width+x] + (1-fy)*newData[(y+(uy-ly))*width+x];
        }
    }
    return data;
}

exports.verticalDilation = function (data, width, height, lengthFrac, skewOffset) {
    if (skewOffset===undefined) skewOffset = 0;
    var x, y, c, lx, ux, fx, sv;
    var radius = Math.floor(lengthFrac/2);
    var queue = new MorphologicalQueue(2*radius+1, greaterThanOrEqualTo);
    var skewedBeginX = Math.min(0,Math.floor((height-1)*skewOffset));
    var skewedEndX = width + Math.max(0,Math.ceil((height-1)*skewOffset));
    var newData = new Float32Array((width+1)*height);
    for(x=skewedBeginX; x<skewedEndX; x++) {
        queue.empty();
        for(y=0; y<height+radius; y++) {
            fx = x - y*skewOffset;
            lx = Math.floor(fx);
            ux = Math.ceil(fx);
            fx -= lx;
            if (y<height && 0<=ux && lx<width) {
                if (lx<0 || ux==lx) {
                    sv = data[y*width+ux];
                } else if (ux==width) {
                    sv = data[y*width+lx];
                } else {
                    sv = data[y*width+lx] + fx*(data[y*width+ux]-data[y*width+lx]);
                }
                queue.enqueue(y, sv);
            } else {
                queue.enqueue(y);
            }
            if (y>=radius) {
                fx = x - (y-radius)*skewOffset;
                lx = Math.floor(fx);
                ux = Math.ceil(fx);
                fx -= lx;
                if (0<=ux && lx<width) {
                    newData[(y-radius)*(width+1)+ux] = queue.firstVal();
                }
            }
        }
    }
    for(y=0; y<height; y++) {
        for(x=0; x<width; x++) {
            fx = x - y*skewOffset;
            lx = Math.floor(fx);
            ux = Math.ceil(fx);
            fx -= lx;
            data[y*width+x] = fx*newData[y*(width+1)+x] + (1-fx)*newData[y*(width+1)+x+(ux-lx)];
        }
    }
    return data;
}

exports.horizontalErosion = function (data, width, height, lengthFrac, skewOffset) {
    if (skewOffset===undefined) skewOffset = 0;
    var x, y, c, ly, uy, fy, sv;
    var radius = Math.floor(lengthFrac/2);
    var queue = new MorphologicalQueue(2*radius+1, lessThanOrEqualTo);
    var skewedBeginY = Math.min(0,Math.floor((width-1)*skewOffset));
    var skewedEndY = height + Math.max(0,Math.ceil((width-1)*skewOffset));
    var newData = new Float32Array(width*(height+1));
    for(y=skewedBeginY; y<skewedEndY; y++) {
        queue.empty();
        for(x=0; x<width+radius; x++) {
            fy = y - x*skewOffset;
            ly = Math.floor(fy);
            uy = Math.ceil(fy);
            fy -= ly;
            if (x<width && 0<=uy && ly<height) {
                if (ly<0 || uy==ly) {
                    sv = data[uy*width+x];
                } else if (uy==height) {
                    sv = data[ly*width+x];
                } else {
                    sv = data[ly*width+x] + fy*(data[uy*width+x]-data[ly*width+x]);
                }
                queue.enqueue(x, sv);
            } else {
                queue.enqueue(x);
            }
            if (x>=radius) {
                fy = y - (x-radius)*skewOffset;
                ly = Math.floor(fy);
                uy = Math.ceil(fy);
                fy -= ly;
                if (0<=uy && ly<height) {
                    newData[uy*width+x-radius] = queue.firstVal();
                }
            }
        }
    }
    for(y=0; y<height; y++) {
        for(x=0; x<width; x++) {
            fy = y - x*skewOffset;
            ly = Math.floor(fy);
            uy = Math.ceil(fy);
            fy -= ly;
            data[y*width+x] = fy*newData[y*width+x] + (1-fy)*newData[(y+(uy-ly))*width+x];
        }
    }
    return data;
}

exports.verticalErosion = function (data, width, height, lengthFrac, skewOffset) {
    if (skewOffset===undefined) skewOffset = 0;
    var x, y, c, lx, ux, fx, sv;
    var radius = Math.floor(lengthFrac/2);
    var queue = new MorphologicalQueue(2*radius+1, lessThanOrEqualTo);
    var skewedBeginX = Math.min(0,Math.floor((height-1)*skewOffset));
    var skewedEndX = width + Math.max(0,Math.ceil((height-1)*skewOffset));
    var newData = new Float32Array((width+1)*height);
    for(x=skewedBeginX; x<skewedEndX; x++) {
        queue.empty();
        for(y=0; y<height+radius; y++) {
            fx = x - y*skewOffset;
            lx = Math.floor(fx);
            ux = Math.ceil(fx);
            fx -= lx;
            if (y<height && 0<=ux && lx<width) {
                if (lx<0 || ux==lx) {
                    sv = data[y*width+ux];
                } else if (ux==width) {
                    sv = data[y*width+lx];
                } else {
                    sv = data[y*width+lx] + fx*(data[y*width+ux]-data[y*width+lx]);
                }
                queue.enqueue(y, sv);
            } else {
                queue.enqueue(y);
            }
            if (y>=radius) {
                fx = x - (y-radius)*skewOffset;
                lx = Math.floor(fx);
                ux = Math.ceil(fx);
                fx -= lx;
                if (0<=ux && lx<width) {
                    newData[(y-radius)*(width+1)+ux] = queue.firstVal();
                }
            }
        }
    }
    for(y=0; y<height; y++) {
        for(x=0; x<width; x++) {
            fx = x - y*skewOffset;
            lx = Math.floor(fx);
            ux = Math.ceil(fx);
            fx -= lx;
            data[y*width+x] = fx*newData[y*(width+1)+x] + (1-fx)*newData[y*(width+1)+x+(ux-lx)];
        }
    }
    return data;
}

exports.horizontalClosing = function (data, width, height, lengthFrac, skewOffset) {
    if (skewOffset===undefined) skewOffset = 0;
    var x, y, c, ly, uy, fy, sv;
    var length = Math.floor(lengthFrac); // The actual integer length is length+1 (length is the number of "spacings" between the length+1 values that we consider).
    var queue = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
    var queue2 = new MorphologicalQueue(length+1, lessThanOrEqualTo);
    var skewedBeginY = Math.min(0,Math.floor((width-1)*skewOffset));
    var skewedEndY = height + Math.max(0,Math.ceil((width-1)*skewOffset));
    var newData = new Float32Array(width*(height+1));
    for(y=skewedBeginY; y<skewedEndY; y++) {
        queue.empty();
        queue2.empty();
        for(x=0; x<width+length; x++) {
            fy = y - x*skewOffset;
            ly = Math.floor(fy);
            uy = Math.ceil(fy);
            fy -= ly;
            if (x<width && 0<=uy && ly<height) {
                if (ly<0 || uy==ly) {
                    sv = data[uy*width+x];
                } else if (uy==height) {
                    sv = data[ly*width+x];
                } else {
                    sv = data[ly*width+x] + fy*(data[uy*width+x]-data[ly*width+x]);
                }
                queue.enqueue(x, sv);
            } else {
                queue.enqueue(x);
            }
            queue2.enqueue(x, queue.firstVal()); // Start outputting to the second queue immediately, as the dilation might output values relevant to the erosion that normally do not actually show up in the output.
            if (x>=length) {
                fy = y - (x-length)*skewOffset;
                ly = Math.floor(fy);
                uy = Math.ceil(fy);
                fy -= ly;
                if (0<=uy && ly<height) {
                    newData[uy*width+x-length] = queue2.firstVal();
                }
            }
        }
    }
    for(y=0; y<height; y++) {
        for(x=0; x<width; x++) {
            fy = y - x*skewOffset;
            ly = Math.floor(fy);
            uy = Math.ceil(fy);
            fy -= ly;
            data[y*width+x] = fy*newData[y*width+x] + (1-fy)*newData[(y+(uy-ly))*width+x];
        }
    }
    return data;
}

exports.verticalClosing = function (data, width, height, lengthFrac, skewOffset) {
    if (skewOffset===undefined) skewOffset = 0;
    var x, y, c, lx, ux, fx, sv;
    var length = Math.floor(lengthFrac); // The actual integer length is length+1 (length is the number of "spacings" between the length+1 values that we consider).
    var queue = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
    var queue2 = new MorphologicalQueue(length+1, lessThanOrEqualTo);
    var skewedBeginX = Math.min(0,Math.floor((height-1)*skewOffset));
    var skewedEndX = width + Math.max(0,Math.ceil((height-1)*skewOffset));
    var newData = new Float32Array((width+1)*height);
    for(x=skewedBeginX; x<skewedEndX; x++) {
        queue.empty();
        queue2.empty();
        for(y=0; y<height+length; y++) {
            fx = x - y*skewOffset;
            lx = Math.floor(fx);
            ux = Math.ceil(fx);
            fx -= lx;
            if (y<height && 0<=ux && lx<width) {
                if (lx<0 || ux==lx) {
                    sv = data[y*width+ux];
                } else if (ux==width) {
                    sv = data[y*width+lx];
                } else {
                    sv = data[y*width+lx] + fx*(data[y*width+ux]-data[y*width+lx]);
                }
                queue.enqueue(y, sv);
            } else {
                queue.enqueue(y);
            }
            queue2.enqueue(y, queue.firstVal()); // Start outputting to the second queue immediately, as the dilation might output values relevant to the erosion that normally do not actually show up in the output.
            if (y>=length) {
                fx = x - (y-length)*skewOffset;
                lx = Math.floor(fx);
                ux = Math.ceil(fx);
                fx -= lx;
                if (0<=ux && lx<width) {
                    newData[(y-length)*(width+1)+ux] = queue2.firstVal();
                }
            }
        }
    }
    for(y=0; y<height; y++) {
        for(x=0; x<width; x++) {
            fx = x - y*skewOffset;
            lx = Math.floor(fx);
            ux = Math.ceil(fx);
            fx -= lx;
            data[y*width+x] = fx*newData[y*(width+1)+x] + (1-fx)*newData[y*(width+1)+x+(ux-lx)];
        }
    }
    return data;
}

exports.horizontalOpening = function (data, width, height, lengthFrac, skewOffset) {
    if (skewOffset===undefined) skewOffset = 0;
    var x, y, c, ly, uy, fy, sv;
    var length = Math.floor(lengthFrac); // The actual integer length is length+1 (length is the number of "spacings" between the length+1 values that we consider).
    var queue = new MorphologicalQueue(length+1, lessThanOrEqualTo);
    var queue2 = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
    var skewedBeginY = Math.min(0,Math.floor((width-1)*skewOffset));
    var skewedEndY = height + Math.max(0,Math.ceil((width-1)*skewOffset));
    var newData = new Float32Array(width*(height+1));
    for(y=skewedBeginY; y<skewedEndY; y++) {
        queue.empty();
        queue2.empty();
        for(x=0; x<width+length; x++) {
            fy = y - x*skewOffset;
            ly = Math.floor(fy);
            uy = Math.ceil(fy);
            fy -= ly;
            if (x<width && 0<=uy && ly<height) {
                if (ly<0 || uy==ly) {
                    sv = data[uy*width+x];
                } else if (uy==height) {
                    sv = data[ly*width+x];
                } else {
                    sv = data[ly*width+x] + fy*(data[uy*width+x]-data[ly*width+x]);
                }
                queue.enqueue(x, sv);
            } else {
                queue.enqueue(x);
            }
            queue2.enqueue(x, queue.firstVal()); // Start outputting to the second queue immediately, as the dilation might output values relevant to the erosion that normally do not actually show up in the output.
            if (x>=length) {
                fy = y - (x-length)*skewOffset;
                ly = Math.floor(fy);
                uy = Math.ceil(fy);
                fy -= ly;
                if (0<=uy && ly<height) {
                    newData[uy*width+x-length] = queue2.firstVal();
                }
            }
        }
    }
    for(y=0; y<height; y++) {
        for(x=0; x<width; x++) {
            fy = y - x*skewOffset;
            ly = Math.floor(fy);
            uy = Math.ceil(fy);
            fy -= ly;
            data[y*width+x] = fy*newData[y*width+x] + (1-fy)*newData[(y+(uy-ly))*width+x];
        }
    }
    return data;
}

exports.verticalOpening = function (data, width, height, lengthFrac, skewOffset) {
    if (skewOffset===undefined) skewOffset = 0;
    var x, y, c, lx, ux, fx, sv;
    var length = Math.floor(lengthFrac); // The actual integer length is length+1 (length is the number of "spacings" between the length+1 values that we consider).
    var queue = new MorphologicalQueue(length+1, lessThanOrEqualTo);
    var queue2 = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
    var skewedBeginX = Math.min(0,Math.floor((height-1)*skewOffset));
    var skewedEndX = width + Math.max(0,Math.ceil((height-1)*skewOffset));
    var newData = new Float32Array((width+1)*height);
    for(x=skewedBeginX; x<skewedEndX; x++) {
        queue.empty();
        queue2.empty();
        for(y=0; y<height+length; y++) {
            fx = x - y*skewOffset;
            lx = Math.floor(fx);
            ux = Math.ceil(fx);
            fx -= lx;
            if (y<height && 0<=ux && lx<width) {
                if (lx<0 || ux==lx) {
                    sv = data[y*width+ux];
                } else if (ux==width) {
                    sv = data[y*width+lx];
                } else {
                    sv = data[y*width+lx] + fx*(data[y*width+ux]-data[y*width+lx]);
                }
                queue.enqueue(y, sv);
            } else {
                queue.enqueue(y);
            }
            queue2.enqueue(y, queue.firstVal()); // Start outputting to the second queue immediately, as the dilation might output values relevant to the erosion that normally do not actually show up in the output.
            if (y>=length) {
                fx = x - (y-length)*skewOffset;
                lx = Math.floor(fx);
                ux = Math.ceil(fx);
                fx -= lx;
                if (0<=ux && lx<width) {
                    newData[(y-length)*(width+1)+ux] = queue2.firstVal();
                }
            }
        }
    }
    for(y=0; y<height; y++) {
        for(x=0; x<width; x++) {
            fx = x - y*skewOffset;
            lx = Math.floor(fx);
            ux = Math.ceil(fx);
            fx -= lx;
            data[y*width+x] = fx*newData[y*(width+1)+x] + (1-fx)*newData[y*(width+1)+x+(ux-lx)];
        }
    }
    return data;
}

// Implements a closing followed by an opening
exports.horizontalOpenClose = function (data, width, height, lengthFrac, skewOffset) {
    if (skewOffset===undefined) skewOffset = 0;
    var x, y, c, ly, uy, fy, sv;
    var length = Math.floor(lengthFrac);
    var queue = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
    var queue2 = new MorphologicalQueue(2*length+1, lessThanOrEqualTo); // Combines the erosion of the close with the erosion of the open.
    var queue3 = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
    var skewedBeginY = Math.min(0,Math.floor((width-1)*skewOffset));
    var skewedEndY = height + Math.max(0,Math.ceil((width-1)*skewOffset));
    var skewedBeginX, skewedEndX;
    var newData = new Float32Array(width*(height+1));
    for(y=skewedBeginY; y<skewedEndY; y++) {
        queue.empty();
        queue2.empty();
        queue3.empty();
        // We process all positions that are potentially affected by input pixels through (linear) interpolation.
        // If skewOffset > 0:
        // floor(y - x*skewOffset) < height
        //        y - x*skewOffset < height
        //              y - height < x*skewOffset
        //                       x > (y - height)/skewOffset
        // ceil(y - x*skewOffset) => 0
        //       y - x*skewOffset  > -1
        //                  y + 1  > x*skewOffset
        //                      x <  (y + 1)/skewOffset
        // If skewOffset < 0 then both inequalities are reversed in the last steps.
        // Note that there is a slight change this goes wrong in terms of rounding.
        //   If rounding ever proves a problem, we can easily precompute all offsets (-x*skewOffset for all x) and update skewBeginX/skewEndX for each row.
        //   Since both the begin and end should change monotonously this only adds O(width) steps in TOTAL (for all rows).
        if (skewOffset>0) {
            skewedBeginX = Math.max(0, Math.floor((y - height)/skewOffset)+1); // Largest integer <= (y - height)/skewOffset plus one, so actually smallest integer > (y - height)/skewOffset, and this is where x starts.
            skewedEndX = Math.min(width, Math.ceil((y + 1)/skewOffset)); // Smallest integer >= (y + 1)/skewOffset, and x remains an integer strictly less than this, so less than (y + 1)/skewOffset.
        } else if (skewOffset==0) {
            skewedBeginX = 0;
            skewedEndX = width;
        } else {
            skewedBeginX = Math.max(0, Math.floor((y + 1)/skewOffset)+1); // Smallest integer > (y + 1)/skewOffset, and x starts there.
            skewedEndX = Math.min(width, Math.ceil((y - height)/skewOffset)); // Smallest integer >= (y - height)/skewOffset, and x stays strictly smaller.
        }
        // This loop starts at the first pixel on this (skewed) row that might be affected and ends with the last pixel that has an output affected by an input pixel.
        // In other words, before the position skewedBeginX all queues would be empty, so we can just as well start at skewedBeginX, and there is little point in continuing past skewedEndX, as the queues would be empty again.
        for(x=skewedBeginX; x<skewedEndX+2*length; x++) {
            if (x<skewedEndX) { // As long as our position is affected by the input
                // x and y are in the skewed coordinate system, ly, uy and fy describe the lower bound, upper bound and fractional part of the original coordinate.
                fy = y - x*skewOffset;
                ly = Math.floor(fy);
                uy = Math.ceil(fy);
                fy -= ly;
                if (ly<0 || uy==ly) { // Clamping
                    sv = data[uy*width+x];
                } else if (uy==height) { // Clamping
                    sv = data[ly*width+x];
                } else { // Linear interpolation
                    sv = lerp(data[ly*width+x], data[uy*width+x], fy);
                }
                queue.enqueue(x, sv);
            } else {
                queue.enqueue(x); // This is important to keep evicting values that are too old.
            }
            queue2.enqueue(x, queue.firstVal()); // Start outputting to the second queue immediately, as the dilation might output values relevant to the erosion that normally do not actually show up in the output.
            queue3.enqueue(x, queue2.firstVal());
            if (x>=skewedBeginX+2*length) { // As soon as the output is in the range of the input.
                // Conceptually, we want to output to (x,ly+fy), taking the same definitions as above, unfortunately that is not necessarily at a grid position, and we don't even know all neighbouring data yet, so we cannot interpolate at this point.
                // Instead, we do know that for this x, the neighbouring y-positions are +- 1, so we essentially want to write to a grid shifted by fy-ly.
                // So we can use either the floor or ceiling of (ly+fy) as the integer y coordinate.
                // It is most convenient to use the ceiling (uy), as this is guaranteed to be >= 0 (because of skewedBeginX/skewedEndX).
                uy = Math.ceil(y - (x-2*length)*skewOffset);
                newData[uy*width+(x-2*length)] = queue3.firstVal();
            }
        }
    }
    for(y=0; y<height; y++) {
        for(x=0; x<width; x++) {
            // The y-positions in the array newData are the ceiling of the actual positions of the values.
            // So the value at (u,v) in newData is actually the value at position (u,v-(1-fy)).
            // To sample at (x,y), we thus need to sample at (x,y+(1-fy)) in newData.
            // Note that if fy==0, we should sample at (x,y), not (x,y+1).
            // In essence, fy should be in the range (0,1] here, rather than [0,1).
            // This is accomplished by actually computing 1-fy
            fy = -x*skewOffset;
            fy = Math.ceil(fy) - fy; // This actually computes 1-fy, but if fy was an integer, we have 1-fy==0, rather than 1-fy==1.
            data[y*width+x] = lerp(newData[y*width+x], newData[(y+1)*width+x], fy);
            // Finally, it is not too difficult to see that all values used here for interpolation (with non-zero weights) are valid.
            // This stems from the fact that the filter considers all y-positions whose upper or lower bound are within the domain of the original image.
            // In other words, if (x,y) is a position in the output and fy!=0, there are always positions in newData above and below it, if fy=0, it is even more trivial.
        }
    }
    return data;
}

exports.verticalOpenClose = function (data, width, height, lengthFrac, skewOffset) {
    // For comments, see the horizontal version.
    if (skewOffset===undefined) skewOffset = 0;
    var x, y, c, lx, ux, fx, sv;
    var length = Math.floor(lengthFrac);
    var queue = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
    var queue2 = new MorphologicalQueue(2*length+1, lessThanOrEqualTo);
    var queue3 = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
    var skewedBeginX = Math.min(0,Math.floor((height-1)*skewOffset));
    var skewedEndX = width + Math.max(0,Math.ceil((height-1)*skewOffset));
    var skewedBeginY, skewedEndY;
    var newData = new Float32Array((width+1)*height);
    for(x=skewedBeginX; x<skewedEndX; x++) {
        queue.empty();
        queue2.empty();
        queue3.empty();
        if (skewOffset>0) {
            skewedBeginY = Math.max(0, Math.floor((x - width)/skewOffset)+1);
            skewedEndY = Math.min(height, Math.ceil((x + 1)/skewOffset));
        } else if (skewOffset==0) {
            skewedBeginY = 0;
            skewedEndY = height;
        } else {
            skewedBeginY = Math.max(0, Math.floor((x + 1)/skewOffset)+1);
            skewedEndY = Math.min(height, Math.ceil((x - width)/skewOffset));
        }
        for(y=skewedBeginY; y<skewedEndY+2*length; y++) {
            if (y<skewedEndY) {
                fx = x - y*skewOffset;
                lx = Math.floor(fx);
                ux = Math.ceil(fx);
                fx -= lx;
                if (lx<0 || ux==lx) {
                    sv = data[y*width+ux];
                } else if (ux==width) {
                    sv = data[y*width+lx];
                } else {
                    sv = lerp(data[y*width+lx], data[y*width+ux], fx);
                }
                queue.enqueue(y, sv);
            } else {
                queue.enqueue(y);
            }
            queue2.enqueue(y, queue.firstVal());
            queue3.enqueue(y, queue2.firstVal());
            if (y>=skewedBeginY+2*length) {
                ux = Math.ceil(x - (y-2*length)*skewOffset);
                newData[(y-2*length)*(width+1)+ux] = queue3.firstVal();
            }
        }
    }
    for(y=0; y<height; y++) {
        for(x=0; x<width; x++) {
            fx = -y*skewOffset;
            fx = Math.ceil(fx) - fx;
            data[y*width+x] = lerp(newData[y*(width+1)+x], newData[y*(width+1)+x+1], fx);
        }
    }
    return data;
}

// Implements an opening followed by a closing
exports.horizontalCloseOpen = function (data, width, height, lengthFrac, skewOffset) {
    if (skewOffset===undefined) skewOffset = 0;
    var x, y, c, ly, uy, fy, sv;
    var length = Math.floor(lengthFrac);
    var queue = new MorphologicalQueue(length+1, lessThanOrEqualTo);
    var queue2 = new MorphologicalQueue(2*length+1, greaterThanOrEqualTo);
    var queue3 = new MorphologicalQueue(length+1, lessThanOrEqualTo);
    var skewedBeginY = Math.min(0,Math.floor((width-1)*skewOffset));
    var skewedEndY = height + Math.max(0,Math.ceil((width-1)*skewOffset));
    var skewedBeginX, skewedEndX;
    var newData = new Float32Array(width*(height+1));
    for(y=skewedBeginY; y<skewedEndY; y++) {
        queue.empty();
        queue2.empty();
        queue3.empty();
        if (skewOffset>0) {
            skewedBeginX = Math.max(0, Math.floor((y - height)/skewOffset)+1);
            skewedEndX = Math.min(width, Math.ceil((y + 1)/skewOffset));
        } else if (skewOffset==0) {
            skewedBeginX = 0;
            skewedEndX = width;
        } else {
            skewedBeginX = Math.max(0, Math.floor((y + 1)/skewOffset)+1);
            skewedEndX = Math.min(width, Math.ceil((y - height)/skewOffset));
        }
        for(x=skewedBeginX; x<skewedEndX+2*length; x++) {
            if (x<skewedEndX) {
                fy = y - x*skewOffset;
                ly = Math.floor(fy);
                uy = Math.ceil(fy);
                fy -= ly;
                if (ly<0 || uy==ly) {
                    sv = data[uy*width+x];
                } else if (uy==height) {
                    sv = data[ly*width+x];
                } else {
                    sv = lerp(data[ly*width+x], data[uy*width+x], fy);
                }
                queue.enqueue(x, sv);
            } else {
                queue.enqueue(x);
            }
            queue2.enqueue(x, queue.firstVal());
            queue3.enqueue(x, queue2.firstVal());
            if (x>=skewedBeginX+2*length) {
                uy = Math.ceil(y - (x-2*length)*skewOffset);
                newData[uy*width+(x-2*length)] = queue3.firstVal();
            }
        }
    }
    for(y=0; y<height; y++) {
        for(x=0; x<width; x++) {
            fy = -x*skewOffset;
            fy = Math.ceil(fy) - fy;
            data[y*width+x] = lerp(newData[y*width+x], newData[(y+1)*width+x], fy);
        }
    }
    return data;
}

exports.verticalCloseOpen = function (data, width, height, lengthFrac, skewOffset) {
    // For comments, see the horizontal version.
    if (skewOffset===undefined) skewOffset = 0;
    var x, y, c, lx, ux, fx, sv;
    var length = Math.floor(lengthFrac);
    var queue = new MorphologicalQueue(length+1, lessThanOrEqualTo);
    var queue2 = new MorphologicalQueue(2*length+1, greaterThanOrEqualTo);
    var queue3 = new MorphologicalQueue(length+1, lessThanOrEqualTo);
    var skewedBeginX = Math.min(0,Math.floor((height-1)*skewOffset));
    var skewedEndX = width + Math.max(0,Math.ceil((height-1)*skewOffset));
    var skewedBeginY, skewedEndY;
    var newData = new Float32Array((width+1)*height);
    for(x=skewedBeginX; x<skewedEndX; x++) {
        queue.empty();
        queue2.empty();
        queue3.empty();
        if (skewOffset>0) {
            skewedBeginY = Math.max(0, Math.floor((x - width)/skewOffset)+1);
            skewedEndY = Math.min(height, Math.ceil((x + 1)/skewOffset));
        } else if (skewOffset==0) {
            skewedBeginY = 0;
            skewedEndY = height;
        } else {
            skewedBeginY = Math.max(0, Math.floor((x + 1)/skewOffset)+1);
            skewedEndY = Math.min(height, Math.ceil((x - width)/skewOffset));
        }
        for(y=skewedBeginY; y<skewedEndY+2*length; y++) {
            if (y<skewedEndY) {
                fx = x - y*skewOffset;
                lx = Math.floor(fx);
                ux = Math.ceil(fx);
                fx -= lx;
                if (lx<0 || ux==lx) {
                    sv = data[y*width+ux];
                } else if (ux==width) {
                    sv = data[y*width+lx];
                } else {
                    sv = lerp(data[y*width+lx], data[y*width+ux], fx);
                }
                queue.enqueue(y, sv);
            } else {
                queue.enqueue(y);
            }
            queue2.enqueue(y, queue.firstVal());
            queue3.enqueue(y, queue2.firstVal());
            if (y>=skewedBeginY+2*length) {
                ux = Math.ceil(x - (y-2*length)*skewOffset);
                newData[(y-2*length)*(width+1)+ux] = queue3.firstVal();
            }
        }
    }
    for(y=0; y<height; y++) {
        for(x=0; x<width; x++) {
            fx = -y*skewOffset;
            fx = Math.ceil(fx) - fx;
            data[y*width+x] = lerp(newData[y*(width+1)+x], newData[y*(width+1)+x+1], fx);
        }
    }
    return data;
}

exports.erosionSquare = function (data, width, height, lengthFrac) {
    var x, y;
    var radius = Math.floor(lengthFrac/2);
    var xqueue = new MorphologicalQueue(2*radius+1, lessThanOrEqualTo);
    var yqueues = new Array(width+radius);
    for(x=0; x<yqueues.length; x++) {
        yqueues[x] = new MorphologicalQueue(2*radius+1, lessThanOrEqualTo);
    }
    for(y=0; y<height+radius; y++) {
        xqueue.empty();
        for(x=0; x<width+radius; x++) {
            if (y<height && x<width) {
                xqueue.enqueue(x, data[y*width+x]);
            } else {
                xqueue.enqueue(x);
            }
            yqueues[x].enqueue(y, xqueue.firstVal());
            if (y>=radius && x>=radius) {
                data[(y-radius)*width+(x-radius)] = yqueues[x].firstVal();
            }
        }
    }
    return data;
}

exports.dilationSquare = function (data, width, height, lengthFrac) {
    var x, y;
    var radius = Math.floor(lengthFrac/2);
    var xqueue = new MorphologicalQueue(2*radius+1, greaterThanOrEqualTo);
    var yqueues = new Array(width+radius);
    for(x=0; x<yqueues.length; x++) {
        yqueues[x] = new MorphologicalQueue(2*radius+1, greaterThanOrEqualTo);
    }
    for(y=0; y<height+radius; y++) {
        xqueue.empty();
        for(x=0; x<width+radius; x++) {
            if (y<height && x<width) {
                xqueue.enqueue(x, data[y*width+x]);
            } else {
                xqueue.enqueue(x);
            }
            yqueues[x].enqueue(y, xqueue.firstVal());
            if (y>=radius && x>=radius) {
                data[(y-radius)*width+(x-radius)] = yqueues[x].firstVal();
            }
        }
    }
    return data;
}

exports.closeSquare = function (data, width, height, lengthFrac) {
    var x, y;
    var length = Math.floor(lengthFrac);
    var xqueue = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
    var xqueue2 = new MorphologicalQueue(length+1, lessThanOrEqualTo);
    var yqueues = new Array(width+length);
    var yqueues2 = new Array(width+length);
    for(x=0; x<yqueues.length; x++) {
        yqueues[x] = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
        yqueues2[x] = new MorphologicalQueue(length+1, lessThanOrEqualTo);
    }
    for(y=0; y<height+length; y++) {
        xqueue.empty();
        xqueue2.empty();
        for(x=0; x<width+length; x++) {
            if (y<height && x<width) {
                xqueue.enqueue(x, data[y*width+x]);
            } else {
                xqueue.enqueue(x);
            }
            yqueues[x].enqueue(y, xqueue.firstVal());
            xqueue2.enqueue(x, yqueues[x].firstVal());
            yqueues2[x].enqueue(y, xqueue2.firstVal());
            if (y>=length && x>=length) {
                data[(y-length)*width+(x-length)] = yqueues2[x].firstVal();
            }
        }
    }
    return data;
}

exports.openSquare = function (data, width, height, lengthFrac) {
    var x, y;
    var length = Math.floor(lengthFrac);
    var xqueue = new MorphologicalQueue(length+1, lessThanOrEqualTo);
    var xqueue2 = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
    var yqueues = new Array(width+length);
    var yqueues2 = new Array(width+length);
    for(x=0; x<yqueues.length; x++) {
        yqueues[x] = new MorphologicalQueue(length+1, lessThanOrEqualTo);
        yqueues2[x] = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
    }
    for(y=0; y<height+length; y++) {
        xqueue.empty();
        xqueue2.empty();
        for(x=0; x<width+length; x++) {
            if (y<height && x<width) {
                xqueue.enqueue(x, data[y*width+x]);
            } else {
                xqueue.enqueue(x);
            }
            yqueues[x].enqueue(y, xqueue.firstVal());
            xqueue2.enqueue(x, yqueues[x].firstVal());
            yqueues2[x].enqueue(y, xqueue2.firstVal());
            if (y>=length && x>=length) {
                data[(y-length)*width+(x-length)] = yqueues2[x].firstVal();
            }
        }
    }
    return data;
}

exports.occoSquare = function (data, width, height, lengthFrac) {
    var x, y;
    var length = Math.floor(lengthFrac);
    var xqueue = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
    var xqueue2 = new MorphologicalQueue(2*length+1, lessThanOrEqualTo); // Combine erosion of the closing with erosion of the opening
    var xqueue3 = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
    var xdqueue = new MorphologicalQueue(length+1, lessThanOrEqualTo);
    var xdqueue2 = new MorphologicalQueue(2*length+1, greaterThanOrEqualTo); // Combine dilation of the erosion with dilation of the closing
    var xdqueue3 = new MorphologicalQueue(length+1, lessThanOrEqualTo);
    var yqueues = new Array(width+2*length);
    var yqueues2 = new Array(width+2*length);
    var yqueues3 = new Array(width+2*length);
    var ydqueues = new Array(width+2*length);
    var ydqueues2 = new Array(width+2*length);
    var ydqueues3 = new Array(width+2*length);
    for(x=0; x<yqueues.length; x++) {
        yqueues[x] = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
        yqueues2[x] = new MorphologicalQueue(2*length+1, lessThanOrEqualTo);
        yqueues3[x] = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
        ydqueues[x] = new MorphologicalQueue(length+1, lessThanOrEqualTo);
        ydqueues2[x] = new MorphologicalQueue(2*length+1, greaterThanOrEqualTo);
        ydqueues3[x] = new MorphologicalQueue(length+1, lessThanOrEqualTo);
    }
    for(y=0; y<height+2*length; y++) {
        xqueue.empty();
        xqueue2.empty();
        xqueue3.empty();
        xdqueue.empty();
        xdqueue2.empty();
        xdqueue3.empty();
        for(x=0; x<width+2*length; x++) {
            if (y<height && x<width) {
                xqueue.enqueue(x, data[y*width+x]);
                xdqueue.enqueue(x, data[y*width+x]);
            } else {
                xqueue.enqueue(x);
                xdqueue.enqueue(x);
            }
            
            yqueues[x].enqueue(y, xqueue.firstVal());
            xqueue2.enqueue(x, yqueues[x].firstVal());
            yqueues2[x].enqueue(y, xqueue2.firstVal());
            xqueue3.enqueue(x, yqueues2[x].firstVal());
            yqueues3[x].enqueue(y, xqueue3.firstVal());
            
            ydqueues[x].enqueue(y, xdqueue.firstVal());
            xdqueue2.enqueue(x, ydqueues[x].firstVal());
            ydqueues2[x].enqueue(y, xdqueue2.firstVal());
            xdqueue3.enqueue(x, ydqueues2[x].firstVal());
            ydqueues3[x].enqueue(y, xdqueue3.firstVal());
            
            if (y>=2*length && x>=2*length) {
                data[(y-2*length)*width+(x-2*length)] = 0.5*(yqueues3[x].firstVal() + ydqueues3[x].firstVal());
            }
        }
    }
    return data;
}

